#include <stdio.h>
#include <string.h>

#include "type.h"

/* Constructors
   Yes, I mostly write c++-code
   */

/* The NAME argument is copied. The type argument is preserved as a pointer. */
struct record_member *
make_record_member(char * name, struct type * type)
{
    struct record_member * ne = 
	(struct record_member *)malloc(sizeof(struct record_member));

    ne->name = strdup(name);
    ne->typ = type;

    return ne;
}

struct type *
make_simple_type(enum Etype typ)
{
    struct type * ne = (struct type *)malloc(sizeof(struct type));

    ne->typ = typ;
    ne->size = 0;
    ne->u.element = 0;

    return ne;
}

struct type *
make_varchar_type(enum Etype typ, unsigned short siz)
{
    struct type * ne = make_simple_type(typ);

    ne->size = siz;

    return ne;
}

struct type *
make_array_type(struct type * typ, unsigned short siz)
{
    struct type * ne = make_simple_type(et_array);

    ne->size = siz;
    ne->u.element = typ;

    return ne;
}

struct type *
make_record_type(struct record_member * rm[])
{
    struct type * ne = make_simple_type(et_record);

    ne->u.members = rm;

    return ne;
}


/* Dump a type.
   The type is dumped as:
	type-tag <comma> 		- enum Etype
	reference-to-variable <comma> 	- void *
	size <comma>			- short size of this field (if varchar)
	arrsize <comma>			- short number of elements in the arr
	offset <comma>			- short offset to the next element
   Where:
   type-tag is one of the simple types or varchar.
   reference-to-variable can be a reference to a struct element.
   arrsize is the size of the array in case of array fetches. Otherwise 0.
   size is the maxsize in case it is a varchar. Otherwise it is the size of
   	the variable (required to do array fetches of records).
 */
void dump_a_simple(FILE * o, const char * name, enum Etype typ, 
		   short varcharsize,
		   unsigned short arrsiz, const char * siz);
void dump_a_record(FILE * o, const char * name, unsigned short arrsiz,
		   struct type * typ, const char * offset);


void
dump_a_type(FILE * o, const char * name, struct type * typ)
{
    if (IS_SIMPLE_TYPE(typ->typ))
    {
	dump_a_simple(o, name, typ->typ, typ->size, 0, 0);
    }
    else if (typ->typ == et_array)
    {
	if (IS_SIMPLE_TYPE(typ->u.element->typ))
	    dump_a_simple(o, name, typ->u.element->typ,
			  typ->u.element->size, typ->size, 0);
	else if (typ->u.element->typ == et_array)
	{
	    abort();		/* Array of array, */
	}
	else if (typ->u.element->typ == et_record)
	{
	    /* Array of records. */
	    dump_a_record(o, name, typ->size, typ->u.element, 0);
	}
	else
	{
	    abort();
	}
    }
    else if (typ->typ == et_record)
    {
	dump_a_record(o, name, 0, typ, 0);
    }
    else
    {
	abort();
    }
}


/* If siz is NULL, then the offset is 0, if not use siz as a
   string, it represents the offset needed if we are in an array of records. */
void
dump_a_simple(FILE * o, const char * name, enum Etype typ,
	      short varcharsize,
	      unsigned short arrsiz,
	      const char * siz)
{
    switch (typ)
    {
    case et_char:
	fprintf(o, "et_char,&%s,0,%d,%s, ", name, arrsiz, 
		siz == NULL ? "sizeof(char)" : siz);
	break;
    case et_unsigned_char:
	fprintf(o, "et_unsigned_char,&%s,0,%d,%s, ", name, arrsiz,
		siz == NULL ? "sizeof(unsigned char)" : siz);
	break;
    case et_short:
	fprintf(o, "et_short,&%s,0,%d,%s, ", name, arrsiz,
		siz == NULL ? "sizeof(short)" : siz);
	break; 
    case et_unsigned_short:
	fprintf(o, 
		"et_unsigned_short,&%s,0,%d,%s, ", name, arrsiz,
		siz == NULL ? "sizeof(unsigned short)" : siz);
	break;
    case et_int:
	fprintf(o, "et_int,&%s,0,%d,%s, ", name, arrsiz,
		siz == NULL ? "sizeof(int)" : siz);
	break;
    case et_unsigned_int:
	fprintf(o, "et_unsigned_int,&%s,0,%d,%s, ", name, arrsiz,
		siz == NULL ? "sizeof(unsigned int)" : siz);
	break;
    case et_long:
	fprintf(o, "et_long,&%s,0,%d,%s, ", name, arrsiz,
		siz == NULL ? "sizeof(long)" : siz);
	break;
    case et_unsigned_long:
	fprintf(o, "et_unsigned_int,&%s,0,%d,%s, ", name, arrsiz,
		siz == NULL ? "sizeof(unsigned int)" : siz);
	break;
    case et_float:
	fprintf(o, "et_float,&%s,0,%d,%s, ", name, arrsiz,
		siz == NULL ? "sizeof(float)" : siz);
	break;
    case et_double:
	fprintf(o, "et_double,&%s,0,%d,%s, ", name, arrsiz,
		siz == NULL ? "sizeof(double)" : siz);
	break;
    case et_varchar:
    case et_varchar2:
	if (siz == NULL)
	    fprintf(o, "et_varchar,&%s,%d,%d,sizeof(struct varchar_%s), ", 
		    name,
		    varcharsize,
		    arrsiz, name);
	else
	    fprintf(o, "et_varchar,&%s,%d,%d,%s, ", 
		    name, 
		    varcharsize,
		    arrsiz, siz);
	break;
    default:
	abort();
    }
}


/* Penetrate a record and dump the contents. */
void
dump_a_record(FILE * o,
	      const char * name, unsigned short arrsiz,
	      struct type * typ, const char * offsetarg)
{
    /* If offset is NULL, then this is the first recursive level. If not then 
       we are in a record in a record and the offset is used as offset.
     */
    struct record_member ** p;
    char obuf[BUFSIZ];
    char buf[BUFSIZ];
    const char * offset;

    if (offsetarg == NULL)
    {
	sprintf(obuf, "sizeof(%s)", name);
	offset = obuf;
    }
    else
    {
	offset = offsetarg;
    }

    for (p = typ->u.members; *p; p++)
    {
	if (IS_SIMPLE_TYPE((*p)->typ->typ))
	{
	    sprintf(buf, "%s.%s", name, (*p)->name);
	    dump_a_simple(o, buf, (*p)->typ->typ, (*p)->typ->size,
			  arrsiz, offset);
	}
	else if ((*p)->typ->typ == et_array)
	{
	    int i;

	    for (i = 0; i < (*p)->typ->size; i++)
	    {
		if (IS_SIMPLE_TYPE((*p)->typ->u.element->typ))
		{
		    sprintf(buf, "%s.%s[%d]", name, (*p)->name, i);
		    dump_a_simple(o, buf, (*p)->typ->typ, (*p)->typ->size,
				  arrsiz, offset);
		}
		else if((*p)->typ->u.element->typ == et_array)
		{
		    /* Array within an array. NOT implemented yet. */
		    abort();
		}
		else if ((*p)->typ->u.element->typ == et_record)
		{
		    /* Record within array within record. NOT implemented yet. */
		    abort();
		}
		else	
		{
		    /* Unknown type */
		    abort();
		}
	    }
	}
	else if ((*p)->typ->typ == et_record)
	{
	    /* Record within a record */
	    sprintf(buf, "%s.%s", name, (*p)->name);
	    dump_a_record(o, buf, arrsiz, (*p)->typ, offset);
	}
	else
	{
	    /* Unknown type */
	    abort();
	}
    }
}


const char *
type_name(enum Etype typ)
{
    switch (typ)
    {
    case et_char: 		return "char";
    case et_unsigned_char:	return "unsigned char";
    case et_short: 		return "short";
    case et_unsigned_short: 	return "unsigned short";
    case et_int:		return "int";
    case et_unsigned_int:	return "unsigned int";
    case et_long:		return "long";
    case et_unsigned_long:	return "unsigned long";
    case et_float:		return "float";
    case et_double:		return "double";
    default:
	abort();
    }
}


/* Freeing is not really that important. Since we throw away the process
   anyway. Lets implement that last! */

void
free_record_member(struct record_member * rm)
{
}

void
free_type(struct type * typ)
{
}
